@@ -4,17 +4,28 @@ use 5.006;
 use strict;
 use warnings;
 
-use Scalar::Util qw/reftype/;
-use Encode qw/decode/;
+use Scalar::Util qw/reftype refaddr/;
+use Encode qw/decode encode/;
 
-our $VERSION = '0.01';
-our (@MAYBE, $caller);
+our $VERSION = '0.02';
+our (%HELP, @MAYBE, $caller, $destruct);
+
+BEGIN {
+	%HELP = (
+		arrayref => sub { return map { $_[0]->($_) } @{ $_[1] } },
+		hashref => sub { map { $caller->can('filter_keys') && $caller->filter_keys($_[1]->{$_}, $_) ? 0 
+			: $destruct ? ($_[0]->($_) => $_[0]->($_[1]->{$_})) : do { $_[1]->{$_} = $_[0]->($_[1]->{$_}) } } keys %{ $_[1] }; },
+		scalarref => sub { $$_[1] =~ m/1|0/ ? $_[1] : $_[0]->($$_[1]) },
+		scalar => sub { eval { $_[1] = $_[0]->($_, $_[1], Encode::FB_CROAK); 1; } and last foreach @MAYBE; $_[1]; }
+	);
+}
 
 sub import {
 	my ($pkg) = shift;
 	return unless my @export = @_;
-	@MAYBE = ref $export[scalar @export - 1] ? @{ pop @export } : 'UTF-8';
-	@export = qw/maybe_decode maybe_encode TO_JSON/ if scalar @export == 1 && $export[0] eq 'all';
+	my $opts = ref $export[scalar @export - 1] ? pop @export : ['UTF-8'];
+	@MAYBE = ref $opts eq 'HASH' ? do { $destruct = $opts->{destuct}; @{ $opts->{type} } } : @{ $opts };
+	@export = qw/maybe_decode maybe_encode/ if scalar @export == 1 && $export[0] eq 'all';
 	$caller = scalar caller();
 	{
 		no strict 'refs';
@@ -23,36 +34,32 @@ sub import {
 }
 
 sub maybe_decode {
-	my $self = shift;
-	my $ref = reftype($self);
-	return do { eval { $self = decode($_, $self, Encode::FB_CROAK); 1; } and last foreach @MAYBE; $self; } if !$ref;
-	_recurse($self, $ref, \&maybe_decode);
+	_maybe(shift, \&decode, \&maybe_decode);
 }
 
 sub maybe_encode {
-	my $self = shift;
-	my $ref = reftype($self);
-	return do { eval { $self = encode($_, $self, Encode::FB_CROAK); 1; } and last foreach @MAYBE; $self; } if !$ref;
-	_recurse($self, $ref, \&maybe_encode);
+	_maybe(shift, \&encode, \&maybe_encode);
+}
+
+sub _maybe {
+	my $ref = reftype($_[0]);
+	return $HELP{scalar}->($_[1], $_[0]) if !$ref;
+	return $destruct ? _d_recurse($_[0], $ref, $_[2]) : _recurse($_[0], $ref, $_[2], $_[3] // {});
 }
 
 sub _recurse {
-	my ($self, $ref, $meth) = @_;
+	return $_[0] if $_[3]->{refaddr $_[0]};
+	$_[3]->{refaddr $_[0]} = 1;
+	$_[1] eq 'SCALAR' ? $HELP{scalarref}->($_[2], $_[0]) : $_[1] eq 'ARRAY' 
+		? $HELP{arrayref}->($_[2], $_[0]) : $_[1] eq 'HASH' ? $HELP{hashref}->($_[2], $_[0]) : '*\o/*';
+	return $_[0];
+}
 
-	return $$self =~ m/1|0/ ? $self : $meth->($$self) if ( $ref eq 'SCALAR' );
-
-	return [
-		map {
-			$meth->($_)
-		} @{ $self }
-	] if ($ref eq 'ARRAY');
-
-	my $output = {};
-	for(keys %$self){
-		next if $caller->can('filter_keys') and $caller->filter_keys($self, $_);
-		$output->{$meth->($_)} = $meth->($self->{$_});
-	}
-	return $output;
+sub _d_recurse {
+	return $HELP{scalarref}->($_[2], $_[0]) if ( $_[1] eq 'SCALAR' );
+	return [ $HELP{arrayref}->($_[2], $_[0]) ] if ($_[1] eq 'ARRAY');
+	return { $HELP{hashref}->($_[2], $_[0]) } if ($_[1] eq 'HASH');
+	return $_[0];
 }
 
 1;
@@ -65,7 +72,7 @@ Struct::WOP - deeply encode/decode a struct
 
 =head1 VERSION
 
-Version 0.01
+Version 0.02
 
 =cut
 
